﻿# A really huge thank to L'In20Cible for showing how to use TE_Effects!!

# =============================================================================
# >> IMPORTS
# =============================================================================
# Python
from binascii  import unhexlify
from configobj import ConfigObj
from path      import path

# EventScripts
import es
import playerlib

# Source-Python-Extensions
import spe


# =============================================================================
# >> ADDON INFORMATION
# =============================================================================
info = es.AddonInfo()
info.name        = 'SPE Effects'
info.version     = '1.0.1'
info.author      = 'Ayuto'
info.url         = 'http://addons.eventscripts.com/addons/user/53757'
info.basename    = 'spe_effects'
info.description = 'Provides 41 new effects by using SPE'

es.set('spe_effects_version', info.version)
es.makepublic('spe_effects_version')


# =============================================================================
# >> GLOBAL VARIABLES & INITIALIZATION
# =============================================================================
BASEPATH = path(__file__).parent
EFFECTS  = ConfigObj(BASEPATH.joinpath('effects.ini'))
POINTERS = ConfigObj(BASEPATH.joinpath('pointers.ini'))

if spe.platform == 'nt':
    sig, pos = POINTERS['g_TESystem']['nt']
    g_TESystem = spe.getLocVal('i', spe.getLocVal('i', spe.getPointer(
        unhexlify(''.join(sig.split())), int(pos))))
else:
    sym = POINTERS['g_TESystem']['linux']
    g_TESystem = spe.getLocVal('i', spe.getLocVal('i', spe.findSymbol(sym)))

spe.parseINI('spe_effects/signatures.ini')


# =============================================================================
# >> FUNCTIONS
# =============================================================================
def createEffect(effect, users, *args):
    '''
    Creates an effect by name.
    '''

    if effect not in EFFECTS:
        raise SPEEffectsError('Effect "%s" does not exist'% effect)

    formatter = lambda x: Vector(*x) if hasattr(x, '__iter__') else x
    args = (IRecipientFilter(getUsers(users)),) + tuple(map(formatter, args))
    mapping = EFFECTS[effect]['mapping']
    if len(mapping) != len(args):
        raise SPEEffectsError('Invalid number of arguments for "' + effect + \
            '". Given: %i, Required: %i'% (len(args), len(mapping)))

    func = findVirtualFunc(g_TESystem, int(EFFECTS[effect]['offset']))
    spe.setCallingConvention('thiscall')
    spe.callFunction(func, 'p'+mapping+')v', ((g_TESystem,) + args))

def findVirtualFunc(pointer, offset):
    if spe.platform != 'nt':
        offset += 1

    return spe.getLocVal('i', spe.getLocVal('i', pointer) + (offset * 4))

def getUsers(users):
    if hasattr(users, '__iter__'):
        return map(int, users)

    return playerlib.getUseridList(str(users))

def _setupEffectFunction(effect):
    function = lambda *args: createEffect(effect, *args)
    function.__doc__ = effect + '(' + EFFECTS[effect]['doc'] + ')'
    function.__name__ = effect
    globals()[effect] = function

# Setups all effects as functions
for effect in EFFECTS:
    _setupEffectFunction(effect)


# =============================================================================
# >> CLASSES
# =============================================================================
class SPEEffectsError(Exception): pass


class IRecipientFilter(int):
    def __new__(cls, users):
        pointer = spe.alloc(40)
        spe.call('RecipientFilterConst', pointer)
        for userid in users:
            if es.exists('userid', userid):
                player = spe.getPlayer(int(userid))
                spe.call('AddRecipient', pointer, player)

        return super(cls, cls).__new__(cls, pointer)

    def __init__(self, users):
        pass

    def __del__(self):
        spe.call('RecipientFilterDest', self)
        spe.dealloc(self)


class Vector(int):
    def __new__(cls, x=0, y=0, z=0):
        pointer = spe.alloc(12)
        spe.setLocVal('f', pointer,     float(x))
        spe.setLocVal('f', pointer + 4, float(y))
        spe.setLocVal('f', pointer + 8, float(z))
        return super(cls, cls).__new__(cls, pointer)

    def __init__(self, x=0, y=0, z=0):
        pass

    def __del__(self):
        spe.dealloc(self)